grammar Lasso9;

options {
k=2; 
backtrack=true; 
memoize=true; 
output=AST;
//language=C;
}
tokens {
	UNUSED;
}
@lexer::members {
int inParens = 0;
}
@lexer::header {
package com.lassosoft.lassolab.internal.core.parsers;
}
@parser::header {
package com.lassosoft.lassolab.internal.core.parsers;
}
@members {

}

program: 
	(UTF8_BOM?)! (SHBANG_LINE?)! expressionList EOF! 
	;
expressionList:	
	semi* (expression semi*)* -> ^(expression)*
	;

expression:
	varDecl
	;
varDecl:
	VAR^ open_paren tag_assign (comma tag_assign_opt)* close_paren
	| VAR^ open_paren tag (comma tag_assign_opt)+ close_paren
	| localDecl
	;
localDecl:
	LOCAL^ open_paren tag_assign (comma tag_assign_opt)* close_paren
	| LOCAL^ open_paren tag (comma tag_assign_opt)+ close_paren
	| define
	;

tag_assign:
	tag ASSIGN expression
	;
tag_assign_opt:
	tag_assign
	| tag
	;

define:
	DEFINE^ tag TARGET define_valid_memsigs associate expression
	| DEFINE^ define_valid_sigs associate expression
	| returnSomething
	;

returnSomething:
	(RETURNHOME|RETURN|YIELDHOME|YIELD)^ (assignment|(open_paren close_paren))?
	| assignment
	;

assignment:
	conditional ((ASSIGN|ASSIGN_PROD|ADD_ASSIGN|SUB_ASSIGN|MUL_ASSIGN|DIV_ASSIGN|MOD_ASSIGN)^ expression)?
	;
conditional:
	or_l (TERN_IF^ expression (TERN_ELSE expression)? semi* )?
	;
or_l:
	and_l (OR^ and_l)*
	;
and_l:
	equality (AND^ equality)*
	;
equality:
	relational ((EQUAL|NOT_EQUAL|EQUAL_STRICT|NOT_EQUAL_STRICT)^ relational)*
	;
relational:
	containment ((LESS_EQUAL|GREATER_EQUAL|LESS|GREATER)^ containment)*
	;
containment:
	additive ((CONTAINS|NOT_CONTAINS)^ additive)*
	;
additive:
	multiplicative ((ADD|SUBTRACT)^ multiplicative)*
	;
multiplicative:
	escapeMember (((MULTIPLY^)|(DIVIDE^)|(MODULUS^)) escapeMember)*
	;
escapeMember:
	unary ESCAPE_MEMBER^ (tag|local|var)
	| unary
	;
	
unary:
	unaryC
	| ADD^ unary
	| SUBTRACT^ unary
	| ADD_PRE^ unary
	| SUB_PRE^ unary
	| NOT^ unary
	| DIVIDE^ tag
	;

unaryC:
	unaryB paramSpec (associate expression)?
	| unaryB 
	;
unaryB:
	unaryA (
		TARGET^ (memberList|string)
		| ADD_PRE^
		| SUB_PRE^
	)*
	;
unaryA:
	constant
	| type
	| trait
	| codeblock
	| capture
	| staticarray
	| query_expression
	| ESCAPE_TAG^ (tag|local|var|parenthesised|ADD|SUBTRACT|MULTIPLY|DIVIDE|MODULUS|EQUAL|LESS|GREATER)
	| lvalue
	;
// root unassignable
constant:
	NULL_L
	| TRUE_L
	| FALSE_L
	| INTEGER
	| DECIMAL
	| SELF_L
	| INHERITED
	| PARAMS
	| string
	| COLONCOLON^ (allOperators|tag|UNDERSCORE)
	| KEYWORD
	| LOCAL_VAR INTEGER
	;
// root assignable
lvalue:
	local
	| var
	| DOT^ string
	| DOTDOT^ string
	| DOT^ memberList
	| DOTDOT^ memberList
	| parenthesised
	| invoke
	;

// etc
string: 
	STRING
	| TICKED_STRING
	;
local: 
	LOCAL_VAR^ tag 
	| LOCAL^ open_paren tag close_paren
	;
var: 
	THREAD_VAR^ tag 
	| VAR^ open_paren tag close_paren
	;
parenthesised: open_paren expression close_paren -> expression;
invoke:
	tag paramSpec (associate expression)?
	| tag (associate expression)?
	;
capture:
	open_block expressionList close_block
	| open_autoblock expressionList close_autoblock
	;
codeblock:
	UNDERSCORE signaturenamedparams paramTypeSpec? associate expression
	;
staticarray:
	open_paren COLON (paramExpr (comma! paramExpr)*)? close_paren
	;
type:
	TYPE open_block! semi* typeBody close_block!
	;
trait:
	TRAIT open_block! semi* traitBody close_block!
	;
// type
typeBody:
	typeMemberPart typeBody*
	| typeDataPart typeBody*
	| typeParentPart typeBody*
	| typeMemberPart typeBody*
	| trait semi* typeBody*
	;
typeDataPart:
	DATA tag (ASSIGN expression)? (comma! tag (ASSIGN expression)?)* semi*
	;
typeParentPart:
	PARENT expression semi*
	;
typeMemberPart:
	(PUBLIC|PROTECTED|PRIVATE)^
		(type_valid_memsigsolo
		|type_valid_memsig associate expression
		|tag associate expression)
		(comma! 
			(type_valid_memsigsolo
			|type_valid_memsig associate expression
			|tag associate expression))* semi*
	;

// trait
traitBody:
	(traitProvidePart|traitRequirePart)* traitImportPart? (traitProvidePart|traitRequirePart)*
	;	
traitImportPart:
	IMPORT^ tag (comma tag)* semi*
	;
traitProvidePart:
	PROVIDE trait_valid_provsig associate expression (comma trait_valid_provsig associate expression)* semi*
	;
traitRequirePart:
	REQUIRE trait_valid_reqsig (comma trait_valid_reqsig)* semi*
	;

// query expressions
query_expression:
	from_clause query_body
	;
from_clause:
	FROM tag IN expression join_clauses?
	;
join_clauses:
	join_clause join_clauses*
	;
join_clause:
	JOIN tag IN expression ON expression EQUALS expression (INTO tag)?
	;
query_body:
	from_let_where_clauses? orderby_clause? select_or_group_clause query_continuation?
	;
from_let_where_clauses:
	from_let_where_clause from_let_where_clauses*
	;
from_let_where_clause:
	from_clause
	| let_clause
	| where_clause
	;
let_clause:
	LET tag ASSIGN expression
	;
where_clause:
	WHERE expression
	;
orderby_clause:
	ORDER BY orderings
	;
orderings:
	ordering (comma ordering)*
	;
ordering:
	expression ordering_direction?
	;
ordering_direction:
	ASCENDING
	| DESCENDING
	;
select_or_group_clause:
	select_clause
	| group_clause
	;
select_clause:
	SELECT expression
	;
group_clause:
	GROUP expression BY expression
	;
query_continuation:
	INTO tag join_clauses? query_body
	;
// memberList
memberList: memberInvoke (RETARGET! memberInvoke)* (RETARGET)?;
memberOp: ASSIGN|ADD_ASSIGN|SUB_ASSIGN|MUL_ASSIGN|DIV_ASSIGN|MOD_ASSIGN ;
memberInvoke: tag memberOp open_paren expression close_paren
	| invoke
	;
// passing params
paramSpec: 
	open_paren paramList? close_paren
	;
paramList: 
	paramExpr (comma! paramExpr)* 
	| COLON expression
	;
paramExpr: assignment
	;

// expected params
paramTypeSpec: COLONCOLON^ tag
	| COLONCOLON^ signature_valid_g_op
	;
paramDesc: tag paramTypeSpec? (ASSIGN expression)? ;

// signature related
signature_plain: tag paramTypeSpec? ;
signature_plain_parens: tag signaturenamedparams paramTypeSpec? ;
signature_g_op_parens: signature_valid_g_op signaturenamedparams paramTypeSpec? ;
signature_m_op_parens: signature_valid_m_op signaturenamedparams paramTypeSpec? ;
signature_mem_op_parens: tag signature_valid_mem_op signaturenamedparams_onedesc paramTypeSpec? ;
signature_mem_op: tag signature_valid_mem_op paramTypeSpec? ;
signature_valid_g_op: LESS|GREATER|EQUAL|ADD|SUBTRACT|MULTIPLY|DIVIDE|MODULUS ;
signature_valid_m_op: ADD|SUBTRACT|MULTIPLY|DIVIDE|MODULUS|ADD_ASSIGN|SUB_ASSIGN|MUL_ASSIGN|DIV_ASSIGN|MOD_ASSIGN ;
signature_valid_mem_op: ASSIGN|ADD_ASSIGN|SUB_ASSIGN|MUL_ASSIGN|DIV_ASSIGN|MOD_ASSIGN ;
signaturenamedparams_onedesc: open_paren paramDesc close_paren ;

signaturenamedparams: 
	open_paren close_paren
	| open_paren paramDesc (comma paramDesc)* (comma DOTDOTDOT tag?)? close_paren
	| open_paren DOTDOTDOT tag? close_paren
	;

// define related
define_valid_sigs:
	signature_plain
	| signature_plain_parens
	| signature_g_op_parens
	;
define_valid_memsigs:
	signature_plain
	| signature_plain_parens
	| signature_m_op_parens
	| signature_mem_op_parens
	;
// member tag sigs
type_valid_memsig:
	signature_plain_parens
	| signature_m_op_parens
	| signature_mem_op_parens
	| signature_plain
	;
type_valid_memsigsolo:
	signature_plain
	;
// trait member sigs
trait_valid_reqsig:
	signature_plain
	|signature_plain_parens
	|signature_m_op_parens
	|signature_mem_op_parens
	|signature_mem_op
	;
trait_valid_provsig:
	signature_plain_parens
	|signature_m_op_parens
	|signature_mem_op_parens
	;

// symbol handling of "soft-colons" (end of lines) and semi-colons
associate:
	ASSOCIATE
	;
open_block:
	OPEN_BLOCK!
	;
open_autoblock:
	OPEN_AUTOBLOCK!
	;
close_block:
	CLOSE_BLOCK!
	;
close_autoblock:
	CLOSE_AUTOBLOCK!
	;
comma:	
	COMMA
	;
open_paren:
	OPEN_PAREN!
	;
close_paren:
	CLOSE_PAREN!
	;
semi: 
	(SEMI
	|EOL)!
	;
tag: 
	TAG|TRUE_L|FALSE_L|NULL_L|SELF_L|INHERITED|PARAMS|DEFINE|LOCAL|VAR|TYPE
	|TRAIT|DATA|PUBLIC|PROTECTED|PRIVATE|IMPORT|REQUIRE|PROVIDE|PARENT
	|RETURNHOME|RETURN|YIELDHOME|YIELD|FROM|IN|SELECT|ORDER|BY|INTO|GROUP|JOIN
	|ASCENDING|DESCENDING|EQUALS|LET|WHERE|ON
	;
allOperators: 
	ADD_ASSIGN|SUB_ASSIGN|MUL_ASSIGN|DIV_ASSIGN|MOD_ASSIGN|OR|AND|EQUAL|EQUAL_STRICT
	|NOT_EQUAL|NOT_EQUAL_STRICT|CONTAINS|NOT_CONTAINS|LESS|LESS_EQUAL|GREATER|GREATER_EQUAL
	|MULTIPLY|DIVIDE|MODULUS|ESCAPE_MEMBER|ESCAPE_TAG|NOT|SUBTRACT|ADD|TARGET 
	;

// --
UNDERSCORE : 	'_' ;
INTEGER :		Ints Exp?|Hex ;
DECIMAL :		(Int* DOT Ints Exp?) | Nan | Infinity ;
fragment
Nan :			'NaN' ;
fragment
Infinity :		'infinity' ;
TRUE_L :		'true' ;
FALSE_L :		'false' ;
NULL_L :		'null' ;
SELF_L :		'self' ;
INHERITED :		'inherited' ;
PARAMS :		'params' ;
DEFINE :		'define' ;
LOCAL :			'local' ;
VAR :			'var' ;
TYPE :			'type' ;
TRAIT :			'trait' ;
DATA :			'data' ;
PUBLIC :		'public' ;
PROTECTED :		'protected' ;
PRIVATE :		'private' ;
IMPORT :		'import' ;
REQUIRE :		'require' ;
PROVIDE :		'provide' ;
PARENT :		'parent' ;
RETURNHOME :	('r'|'R')('e'|'E')('t'|'T')('u'|'U')('r'|'R')('n'|'N')('h'|'H')('o'|'O')('m'|'M')('e'|'E');//'returnhome' ;
RETURN :		'return' ;
YIELDHOME : 	'yieldhome' ;
YIELD :			'yield' ;
FROM :			'from' ;
IN :			'in' ;
WHERE :			'where' ;
SELECT :		'select' ;
JOIN :			'join' ;
ORDER :			'order' ;
BY :			'by' ;
INTO :			'into' ;
ON :			'on' ;
EQUALS :		'equals' ;
LET :			'let' ;
ASCENDING :		'ascending' ;
DESCENDING :	'descending' ;
GROUP :			'group' ;
TICKED_STRING :	
	'`' ( options {greedy=false;} : . )* '`' 
	| ']' ( options {greedy=false;} : . )* '['
	| '?>' ( options {greedy=false;} : . )* '<?lasso'
	;
STRING :		'\'' ( EscapeSequence | ~('\''|'\\') )* '\''
				| '"' ( EscapeSequence | ~('"'|'\\') )* '"' ;
COLONCOLON : 	'::' ;
DOTDOTDOT : 	'...' ;
DOTDOT : 		'..' ;
DOT : 			'.' ;

OPEN_PAREN : 	'(' {++inParens;} ;
CLOSE_PAREN : 	')' {--inParens;} ;
EQUAL_STRICT :	'===' ;
EQUAL :			'==' ;
ASSIGN :		'=' ;
ASSOCIATE :		'=>' ;	
NOT_EQUAL_STRICT :	'!==' ;
NOT_EQUAL :			'!=' ;
NOT_CONTAINS :		'!>>' ;
NOT :				'!' ;
ASSIGN_PROD:		':=' ;
CONTAINS :			'>>' ;
GREATER_EQUAL:		'>=' ;
GREATER :		'>' ;
AND :			'&&' | 'and';
RETARGET :		'&' ;
OR :			'||' | 'or';
TERN_ELSE :		'|' ;
TERN_IF :		'?' ;
LESS_EQUAL :	'<=' ;
LESS :			'<' ;
LOCAL_VAR :		'#' ;
THREAD_VAR :	'$' ;
MUL_ASSIGN :	'*=' ;
MULTIPLY :		'*' ;
DIV_ASSIGN :	'/=' ;
DIVIDE :		'/' ;
MOD_ASSIGN :	'%=' ;
MODULUS :		'%' ;
ESCAPE_MEMBER :	'->\\' ;
TARGET :		'->' ;
ESCAPE_TAG :	'\\' ;
ADD_ASSIGN :	'+=' ;
ADD_PRE :		'++' ;
ADD :			'+' ;
SUB_ASSIGN :	'-=' ;
SUB_PRE :		'--' ;
SUBTRACT :		'-' ;
OPEN_AUTOBLOCK :	'{^' ;
OPEN_BLOCK : 		'{' ;
CLOSE_AUTOBLOCK :	'^}';
CLOSE_BLOCK :		'}';
SEMI : 			';' ;
COLON : 		':' ;
COMMA :			',' ;
COMMENT :		'/*' ( options {greedy=false;} : . )* '*/' {$channel=HIDDEN;} ;
LINE_COMMENT : 	'//' ~('\n'|'\r')* '\r'? '\n' {$channel=HIDDEN;} ;
SHBANG_LINE :  	'#!' ~('\n'|'\r')* ;
KEYWORD :		'-' (TAG|TICKED_STRING) ;
TAG : 			UNDERSCORE?('a'..'z'|'A'..'Z')('a'..'z'|'A'..'Z'|'0'..'9'|UNDERSCORE)* ;
UTF8_BOM :		('\u00EF' '\u00BB' '\u00BF')|('\u00D4' '\u00AA' '\u00F8') ;
WS : 			('\u0009'|'\u000b'|'\u000c'|'\u0020'|'\u00a0'
				|'\u1680'|'\u2000'|'\u2001'|'\u2002'|'\u2003'
				|'\u2004'|'\u2005'|'\u2006'|'\u2007'|'\u2008'
				|'\u2009'|'\u200A'|'\u200B'|'\u202F'|'\u3000') { $channel=HIDDEN; } ;
EOL : 			('\n'|'\r\n'|'\r'|'\u2028'|'\u2029') 
				{if(inParens > 0) { $channel=HIDDEN; };} ;

fragment
Sign : 			('+'|'-') ;
fragment
Int : 			('0'..'9') ;
fragment
Ints : 			Int+ ;
fragment
SignedInt:		Sign? Ints ;
fragment
HexDigit :		('0'..'9'|'a'..'f'|'A'..'F') ;
fragment
Hex : 			'0'('x'|'X')HexDigit+ ;
fragment
Exp : 			('e'|'E')SignedInt ;
fragment
EscapeSequence :	'\\'(('a'|'b'|'t'|'n'|'v'|'f'|'r'|'e'|'"'|'\''|'?'|'\\')
			| ('u' HexDigit HexDigit HexDigit HexDigit)
			| ('U' HexDigit HexDigit HexDigit HexDigit HexDigit HexDigit HexDigit HexDigit)
			| ('x' HexDigit HexDigit?)
			| ('x{' HexDigit (HexDigit (HexDigit (HexDigit (HexDigit (HexDigit (HexDigit HexDigit?)?)?)?)?)?)? '}')
			| ('0'..'7')(('0'..'7')('0'..'7')?)?
			| 'c' (('A'..'Z')|('a'..'z'))
			| ':' (('A'..'Z')|('a'..'z')|' '|'-'|Int) ':'
			| EOL WS*
			) ;

