#!/usr/bin/lasso9

!sys_getenv('LASSO9_RETAIN_COMMENTS')?
	fail('This program requires that Lasso be run with the LASSO9_RETAIN_COMMENTS environment variable set. Run Lasso with LASSO9_RETAIN_COMMENTS set to 1')

define reference_initDb(db::sqlite_db) => {

	#db->executeNow('CREATE TABLE IF NOT EXISTS types_etc (\
														id INTEGER PRIMARY KEY AUTOINCREMENT, \
														type_name TEXT,\
														parent_name TEXT DEFAULT(NULL),\
														role INTEGER, \
														flags INTEGER DEFAULT(0),\
														docs TEXT DEFAULT(NULL))')
	
	#db->executeNow('CREATE TABLE IF NOT EXISTS types_traits_junction (\
														id INTEGER PRIMARY KEY AUTOINCREMENT, \
														type_name TEXT, \
														trait_name TEXT)')
													
	#db->executeNow('CREATE TABLE IF NOT EXISTS methods (\
														id INTEGER PRIMARY KEY AUTOINCREMENT,\
														decorated_name TEXT,\
														method_name TEXT,\
														owner_name INTEGER,\
														returns_name INTEGER,\
														rest_name INTEGER,\
														requirement INTEGER(1),\
														flags INTEGER DEFAULT(0),\
														docs TEXT)')

	#db->executeNow('CREATE TABLE IF NOT EXISTS method_params (\
														id INTEGER PRIMARY KEY AUTOINCREMENT,\
														id_method INTEGER,\
														param_name TEXT,\
														param_type TEXT,\
														param_order INTEGER,\
														flags INTEGER)')

	#db->executeNow('CREATE TABLE IF NOT EXISTS method_comments (\
														id INTEGER PRIMARY KEY AUTOINCREMENT,\
														id_method INTEGER,\
														author_name TEXT,\
														comment TEXT,\
														comment_date TEXT DEFAULT(CURRENT_TIMESTAMP),\
														flags INTEGER)')

	#db->executeNow('CREATE TABLE IF NOT EXISTS type_comments (\
														id INTEGER PRIMARY KEY AUTOINCREMENT,\
														id_type INTEGER,\
														author_name TEXT,\
														comment TEXT,\
														comment_date TEXT DEFAULT(CURRENT_TIMESTAMP),\
														flags INTEGER)')	

	#db->executeNow('CREATE INDEX IF NOT EXISTS types_etc1 on types_etc (type_name,parent_name,role)')
	#db->executeNow('CREATE INDEX IF NOT EXISTS types_traits_junction1 on types_traits_junction (type_name,trait_name)')
	#db->executeNow('CREATE INDEX IF NOT EXISTS methods1 on methods (method_name,owner_name)')
	#db->executeNow('CREATE INDEX IF NOT EXISTS method_params1 on method_params (id_method,param_order)')
	#db->executeNow('CREATE INDEX IF NOT EXISTS method_comments1 on method_comments (id_method)')
	#db->executeNow('CREATE INDEX IF NOT EXISTS type_comments1 on type_comments (id_type)')
														
}

/**!
	Returns the method combined with its parameters
	Used to uniquely identify a method
	Does not include the return type
*/
define reference_makeDecoratedMethodName(method::signature) => {
	local(name = #method->methodName->asString + '(',
		cnt = 0)
	with param in #method->paramDescs
	let methodName = #param->get(1)
	let type = #param->get(2)
	let flags = #param->get(3)
	do {
		#cnt != 0?
			#name->append(',')
			| #cnt = 1
		#flags->bitTest(2)? // keyword
			#name->append('-')
		#name->append(#methodName->asString)
		#type != ::any?
			#name->append(#type->asString)
		#flags->bitTest(1)? // optional
			#name->append('=?')
	}
	return #name + ')'
}

define reference_recordMethod(db, method::signature, req::boolean = false) => {
	stdout('.')
	local(ownerType = #method->typeName->asString,
		decoratedName = reference_makeDecoratedMethodName(#method))
	local(conflictMethod = #db->executeLazy('SELECT id from methods \
											WHERE owner_name '+(#ownerType == '$$unbound$$'? 'ISNULL' | 'LIKE \''+#ownerType->encodeSQL92+'\'')+' \
												AND decorated_name LIKE \''+#decoratedName->encodeSQL92+'\'')->asStaticArray,
		id)
	if (#conflictMethod->size == 0)
		#db->executeNow("INSERT INTO methods (decorated_name, method_name, owner_name, returns_name, rest_name, requirement, docs) \
							VALUES \
							('"+#decoratedName->encodeSQL92+"', \
							'"+#method->methodName->asString->encodeSQL92+"', \
							"+(#ownerType == '$$unbound$$'? 'NULL' | "'"+#ownerType->encodeSQL92+"'")+", '"+#method->returnType->asString->encodeSQL92+"', \
							"+(#method->restName?"'"+#method->restName->asString->encodeSQL92+"'"|'NULL')+", \
							"+(#req?1|0)+", \
							'"+#method->docComment->encodeSQL92+"')")
		#db->executeNow('SELECT last_insert_rowid()') => { #id = #1->get(1) }
	else
		#id = #conflictMethod->get(1)->get(1)
		
		// clear out any existing parameters
		#db->executeNow('DELETE FROM method_params WHERE id_method = ' + #id)
		
		// set the comment regardless
		#method->docComment->size?
			#db->executeNow("UPDATE methods SET docs = '"+#method->docComment->encodeSQL92+"' WHERE id = "+#id)
	/if
	// add records for the parameters
	local(order = 0)
	with param in #method->paramDescs
	do {
		local(stat = 'INSERT INTO method_params (id_method, param_name, param_type, param_order, flags) \
						VALUES \
						('+#id+', \
							\''+#param->get(1)->asString->encodeSQL92+'\', \
							\''+#param->get(2)->asString->encodeSQL92+'\', \
							'+#order+', \
							'+#param->get(3)+')')
		#db->executeNow(#stat)
		#order += 1
	}
}

define reference_recordTrait(db, trait) => {
	stdout('* '+#trait)
	local(conflictTrait = #db->executeLazy('SELECT id, role FROM types_etc WHERE type_name LIKE \'' + #trait->asString->encodeSQL92+'\'')->asStaticArray,
		id, role)
	if (#conflictTrait->size == 0)
		#db->executeNow("INSERT INTO types_etc (type_name, role, docs) \
														VALUES \
														('"+#trait->asString->encodeSQL92+"', 1, \
														'"+#trait->docComment->encodeSQL92+"')")
		#db->executeNow('SELECT last_insert_rowid()') => { #id = #1->get(1) }
	else
		#id = #conflictTrait->get(1)->get(1)
		#role = #conflictTrait->get(1)->get(2)
		1 != integer(#role)?
			fail(-1, 'Tripped up on trait '+#trait->asString)
		#trait->docComment->size?
			#db->executeNow("UPDATE types_etc \
							SET docs = '"+#trait->docComment->encodeSQL92+"' WHERE id = "+#id)
	/if
	
	local(subTraits = set)
	reference_digSubTraits(#trait->getType, #subTraits)
	#subTraits->remove(#trait) // get rid of this trait name
	
	#db->executeNow("DELETE FROM types_traits_junction WHERE type_name = '"+#trait->asString->encodeSQL92+"'")
	
	with subTrait in #subTraits
	do {
		#db->executeNow('INSERT INTO types_traits_junction \
							(type_name,trait_name) \
							VALUES (\''+#trait->asString->encodeSQL92+'\',\''+#subTrait->asString->encodeSQL92+'\')')
	}
	
	with method in #trait->getType->provides
	do reference_recordMethod(#db, #method)
	
	with method in #trait->getType->requires
	do reference_recordMethod(#db, #method, true)
	
	stdoutnl('')
}

define reference_digSubTraitsFor(t::tag, into) => {
	reference_digSubTraits(#t->getType->trait, #into)
	#t != ::null?
		reference_digSubTraitsFor(#t->getType->parent, #into)
}

define reference_digSubTraits(t::void, into) => {} // ?
define reference_digSubTraits(t::trait, into) => {
	#t->name != ::_ && !#t->name->asString->beginsWith('$')?
		#into->insert(#t->name)
	
	with s in #t->subTraits
	do reference_digSubTraits(#s, #into)
}

define reference_recordType(db, type) => {
	stdout('* '+#type)
	local(conflictType = #db->executeLazy('SELECT id, role \
											FROM types_etc \
											WHERE type_name LIKE \'' + #type->asString->encodeSQL92+'\'')->asStaticArray,
		id, role)
	if (#conflictType->size == 0)
		#db->executeNow('INSERT INTO types_etc (type_name, parent_name, role, docs) \
														VALUES \
														(\''+#type->asString->encodeSQL92+'\', \
														\''+#type->getType->parent->asString->encodeSQL92+'\', \
														0, \
														\''+#type->docComment->encodeSQL92+'\')')
		#db->executeNow('SELECT last_insert_rowid()') => { #id = #1->get(1) }
	else
		#id = #conflictType->get(1)->get(1)
		#role = #conflictType->get(1)->get(2)
		0 != integer(#role)?
			fail(-1, 'Tripped up on type '+#type->asString)
		#type->docComment->size?
			#db->executeNow("UPDATE types_etc \
							SET docs = '"+#type->docComment->encodeSQL92+"' WHERE id = "+#id)
	/if	
	
	local(subTraits = set)
	reference_digSubTraitsFor(#type, #subTraits)
	
	#db->executeNow("DELETE FROM types_traits_junction WHERE type_name = '"+#type->asString->encodeSQL92+"'")
	
	with subTrait in #subTraits
	do {
		#db->executeNow('INSERT INTO types_traits_junction \
							(type_name,trait_name) \
							VALUES (\''+#type->asString->encodeSQL92+'\',\''+#subTrait->asString->encodeSQL92+'\')')
	}
	
	with method in #type->getType->listMethods
	where #method->typeName == #type
	where !#method->methodName->asString->beginsWith("'")
	do reference_recordMethod(#db, #method)
	stdoutnl('')
}

local(db = sqlite_db(sys_databasesPath+'lassolab_type_database'))

stdoutnl('Initializing LCAPI')
database_initialize // load lcapi

#db->doWithClose => {

	reference_initDb(#db)
	#db->executeNow('BEGIN TRANSACTION')
	// Load all of the libraries from builtins and lassoserver
	// This forces all possible available types and methods to be registered
	local(srcs = dir(sys_masterHomePath + 'LassoLibraries/')->eachFilePathRecursive)
	
	sys_masterHomePath != sys_homePath and dir(sys_homePath + 'LassoLibraries/')->exists?
		#srcs = tie(#srcs, dir(sys_homePath + 'LassoLibraries/')->eachFilePathRecursive)
	
	with topLevelDir in #srcs
	where #topLevelDir !>> '/.'
	where #topLevelDir !>> '/-'
	where not #topLevelDir->lastComponent->beginsWith('lasso9-core.')
	do protect => {
		handle_error => {
			stdoutnl('Unable to load: ' + #topLevelDir)
		}
		stdoutnl('Attempting to load: ' + #topLevelDir)
		library_thread_loader->loadLibrary(#topLevelDir)
		stdoutnl('Loaded: ' + #topLevelDir)
	}
	
	local(types = sys_listTypes,
		traits = sys_listTraits,
		unboundMethods = sys_listUnboundMethods)

	stdoutnl('Unbound methods')
	// add unbound methods
	with method in #unboundMethods
	do reference_recordMethod(#db, #method)
	
	stdoutnl('\nTraits')
	with trait in #traits
	where !#trait->asString->beginsWith('$')
	do reference_RecordTrait(#db, #trait)
	
	stdoutnl('Types')
	with type in #types
	do reference_recordType(#db, #type)
	
	#db->executeNow('COMMIT TRANSACTION')
} // db close

